/**
 * @param {FC.HumanState} entity
 * @param {Boolean} cheat
 * @returns {HTMLDivElement}
 */
App.Medicine.Modification.eyeSelector = function(entity, cheat = false) {
	const {He, him, his} = getPronouns(entity);

	const player = entity === V.PC;

	let selectedSide = "none";
	let selectedIris = "none";
	let selectedPupil = "none";
	let selectedSclera = "none";

	let removeDiv = document.createElement("div");
	removeDiv.classList.add("choices");
	let applyDiv = document.createElement("div");

	const container = document.createElement("div");
	container.append(
		`${player ? "You have" : `${He} has`} ${App.Desc.eyesColorLong(entity)}, ${hasBothEyes(
			entity) ? "they are" : "it is"} ${App.Desc.eyesType(entity)}.`,
		removeDiv, "You have a number of contact lenses in various colors available. ",
		App.UI.DOM.makeElement("span", `You can change what ${player ? "your" : his} eyes look like.`, "note"),
		assembleLinks(), applyDiv
	);
	updateRemoveDiv();
	updateApplyDiv();
	return container;

	function assembleLinks() {
		const sides = ["left", "right", "both"];
		const irisColors = App.Medicine.Modification.eyeColor.map(color => color.value);
		const pupilShapes = ["none"].concat(App.Medicine.Modification.eyeShape.map(shape => shape.value));
		const scleraColors = ["none"].concat(App.Medicine.Modification.eyeColor.map(color => color.value));
		const div = document.createDocumentFragment();
		div.append(
			assembleList("Side: ", sides, value => selectedSide = value, selectedIris),
			assembleList("Iris: ", irisColors, value => selectedIris = value, selectedSide),
			assembleList("Pupil: ", pupilShapes, value => selectedPupil = value, selectedPupil),
			assembleList("Sclera: ", scleraColors, value => selectedSclera = value, selectedSclera)
		);
		return div;
	}

	/**
	 * @param {string} name
	 * @param {Array<string>} list
	 * @param {function(string):void}callback
	 * @param {string} selected
	 * @returns {HTMLDivElement}
	 */
	function assembleList(name, list, callback, selected) {
		const links = [];

		for (let i = 0; i < list.length; i++) {
			addToggle(list[i], callback, links, list[i] === selected);
		}

		const div = document.createElement("div");
		div.classList.add("choices");
		div.append(name, App.UI.DOM.arrayToList(links, " | ", " | "));
		return div;
	}

	/**
	 * @param {string} value
	 * @param {(value: string) => void} callback
	 * @param {Array<HTMLAnchorElement>} links
	 * @param {boolean} [disabled]
	 */
	function addToggle(value, callback, links, disabled = false) {
		const a = document.createElement("a");
		a.append(capFirstChar(value));
		if (disabled) {
			a.classList.add("disabled");
		}
		a.onclick = () => {
			for (let link of links) {
				link.classList.remove("disabled");
			}
			a.classList.add("disabled");
			callback(value);
			updateRemoveDiv();
			updateApplyDiv();
		};
		links.push(a);
	}

	function updateApplyDiv() {
		$(applyDiv).empty();
		if (selectedSide !== "none" && selectedIris !== "none") {
			// make the following easier to read
			let both = selectedSide === "both";
			let leftGlass = !hasLeftEye(entity) || getLeftEyeType(entity) === 2;
			let rightGlass = !hasRightEye(entity) || getRightEyeType(entity) === 2;

			// base eye
			let r = player ? "" : ` ${him}`;
			if (both) {
				if (leftGlass && rightGlass) {
					r += ` ${selectedIris} glass eyes`;
				} else if (leftGlass || rightGlass) {
					r += ` a glass eye and a ${selectedIris} lens`;
				} else {
					r += ` ${selectedIris} lenses`;
				}
			} else {
				r += " a";
				if ((selectedSide === "left" && leftGlass) || (selectedSide === "right" && rightGlass)) {
					r += ` ${selectedIris} glass eye`;
				} else {
					r += ` ${selectedIris} lens`;
				}
			}
			// pupil & sclera
			if (selectedPupil !== "none" || selectedSclera !== "none") {
				r += " with";
				if (selectedPupil !== "none") {
					r += ` ${both ? selectedPupil : addA(selectedPupil)}`;
					if (both) {
						r += " pupils";
					} else {
						r += " pupil";
					}
					if (selectedSclera !== "none") {
						r += " and";
					}
				}
				if (selectedSclera !== "none") {
					r += ` ${selectedSclera}`;
					if (both) {
						r += " sclerae";
					} else {
						r += " sclera";
					}
				}
			}
			if (!both) {
				r += ` for ${player ? "your" : his} ${selectedSide} eye`;
			}
			r += "?";

			const a = document.createElement("a");
			a.append(player ? "Take" : "Give");
			a.onclick = applyLink;
			applyDiv.append(a, r);
			if (!player) {
				applyDiv.append(" ",
					App.UI.DOM.makeElement("span", "This is independent from eyewear choices.", "note"));
			}
		}
	}

	function applyLink() {
		// make sure the eye exists; give glass eye if there is none
		if ((selectedSide === "left" || selectedSide === "both") && getLeftEyeType(entity) === 0) {
			eyeSurgery(entity, "left", "glass");
		}
		if ((selectedSide === "right" || selectedSide === "both") && getRightEyeType(entity) === 0) {
			eyeSurgery(entity, "right", "glass");
		}

		// apply modifications
		setEyeColorFull(entity, selectedIris,
			selectedPupil === "none" ? "" : selectedPupil,
			selectedSclera === "none" ? "" : selectedSclera,
			selectedSide);
		if (!cheat) {
			cashX(forceNeg(V.modCost), "slaveMod", entity);
		}

		App.UI.reload();
	}

	function updateRemoveDiv() {
		$(removeDiv).empty();
		const links = [];
		let _n = 0;
		// remove lenses
		if (hasLeftEye(entity) && getLeftEyeColor(entity) !== getGeneticEyeColor(entity, "left")) {
			_n++;
			links.push(removeLink("Remove left lens", () => resetEyeColor(entity, "left")));
		}
		if (hasRightEye(entity) && getRightEyeColor(entity) !== getGeneticEyeColor(entity, "right")) {
			_n++;
			links.push(removeLink("Remove right lens", () => resetEyeColor(entity, "right")));
		}
		if (_n === 2) {
			links.push(removeLink("Remove both lenses", () => resetEyeColor(entity, "both")));
		}
		// remove glass eyes
		_n = 0;
		if (getLeftEyeType(entity) === 2) {
			_n++;
			links.push(removeLink("Remove left glass eye", () => eyeSurgery(entity, "left", "remove")));
		}
		if (getRightEyeType(entity) === 2) {
			_n++;
			links.push(removeLink("Remove right glass eye", () => eyeSurgery(entity, "right", "remove")));
		}
		if (_n === 2) {
			links.push(removeLink("Remove both glass eyes", () => eyeSurgery(entity, "both", "remove")));
		}
		if (links.length > 0) {
			removeDiv.append(App.UI.DOM.arrayToList(links, " | ", " | "));
		}
	}

	/**
	 * @param {string} text
	 * @param {() => void} callback
	 */
	function removeLink(text, callback) {
		const a = document.createElement("a");
		a.append(text);
		a.onclick = () => {
			callback();
			App.UI.reload();
		};
		return a;
	}
};

/**
 * Update ears in salon
 * @param {App.Entity.SlaveState} slave
 * @param {object} params
 * @param {number|string} [params.primaryEarColor=0]
 * @param {string} [params.secondaryEarColor=""]
 * @param {Boolean} [params.cheat=false]
 * @returns {HTMLElement}
 */
App.Medicine.Salon.ears = function(slave, {primaryEarColor = 0, secondaryEarColor = "", cheat = false} = {}) {
	let updatePrimary = (newVal) => { primaryEarColor = newVal; apply(); };
	let updateSecondary = (newVal) => { secondaryEarColor = newVal; apply(); };
	const container = document.createElement("div");
	container.id = "salon-ears";
	container.append(content());
	return container;

	function content() {
		const frag = new DocumentFragment();
		if (slave.earT !== "none" && slave.earTColor !== "hairless") {
			const {His, his} = getPronouns(slave);
			let div;
			let p;
			frag.append(`${His} fluffy ears are ${slave.earTColor}.`);

			div = document.createElement("div");
			div.classList.add("choices");
			if (slave.earTColor !== slave.hColor) {
				div.append(
					App.UI.DOM.link(
						"Match current hair",
						() => {
							slave.earTColor = slave.hColor;
							App.Art.refreshSlaveArt(slave, 3, "art-frame");
							apply();
						}
					)
				);
				div.append(" or ");
				App.UI.DOM.appendNewElement("span", div, "choose a new one: ", "note");
			} else {
				App.UI.DOM.appendNewElement("span", div, `Choose a dye color before dyeing ${his} ears:`, "note");
			}
			frag.append(div);

			div = document.createElement("div");
			div.classList.add("choices");
			div.append(`Colors:`);
			div.append(createList(App.Medicine.Modification.Color.Primary, updatePrimary));
			frag.append(div);

			div = document.createElement("div");
			div.classList.add("choices");
			div.append(`Highlights:`);
			div.append(createList(App.Medicine.Modification.Color.Secondary, updateSecondary));
			frag.append(div);

			if (primaryEarColor !== 0) {
				p = document.createElement("p");
				p.classList.add("choices");
				p.append(
					App.UI.DOM.link(
						`Color ${his} ears`,
						() => {
							slave.earTColor = (primaryEarColor + secondaryEarColor);
							App.Art.refreshSlaveArt(slave, 3, "art-frame");
							if (!cheat) {
								cashX(forceNeg(V.modCost), "slaveMod", slave);
							}
							App.Medicine.Salon.ears(slave); // discard selections after locking them in.
						}
					)
				);
				p.append(` ${primaryEarColor}${secondaryEarColor} now?`);
				frag.append(p);
			}
		}
		return frag;
	}

	function createList(array, method) {
		const links = [];
		for (const item of array) {
			const title = item.title || capFirstChar(item.value);
			links.push(
				App.UI.DOM.link(
					title,
					() => method(item.value)
				)
			);
		}
		return App.UI.DOM.generateLinksStrip(links);
	}
	function apply() {
		App.Art.refreshSlaveArt(slave, 3, "art-frame");
		App.Medicine.Salon.ears(
			slave,
			{
				primaryEarColor: primaryEarColor,
				secondaryEarColor: secondaryEarColor,
			}
		);
	}
};

/**
 * Update hair in salon
 * @param {App.Entity.SlaveState} slave
 * @param {object} params
 * @param {number|string} [params.primaryTailColor]
 * @param {string} [params.secondaryTailColor]
 * @param {Boolean} [params.cheat]
 * @returns {JQuery<HTMLElement>}
 */
App.Medicine.Salon.tail = function(slave, {primaryTailColor = 0, secondaryTailColor = "", cheat = false} = {}) {
	const frag = new DocumentFragment();
	let updatePrimary = (newVal) => { primaryTailColor = newVal.value; apply(); };
	let updateSecondary = (newVal) => { secondaryTailColor = newVal.value; apply(); };
	const {His, his} = getPronouns(slave);

	if (slave.tail !== "none") {
		frag.append(tailDye());
	}

	return jQuery("#salon-tail").empty().append(frag);

	function tailDye() {
		const frag = new DocumentFragment();
		let div;
		let p;
		frag.append(`${His} tail is ${slave.tailColor}.`);

		div = document.createElement("div");
		div.classList.add("choices");
		if (slave.origHColor !== slave.hColor) {
			div.append(
				App.UI.DOM.link(
					"Match current hair",
					() => {
						slave.tailColor = slave.hColor;
						App.Art.refreshSlaveArt(slave, 3, "art-frame");
						if (!cheat) {
							cashX(forceNeg(V.modCost), "slaveMod", slave);
						}
						apply();
					}
				)
			);
			div.append(" or ");
			App.UI.DOM.appendNewElement("span", div, "choose a new one: ", "note");
		} else {
			App.UI.DOM.appendNewElement("span", div, `Choose a dye color before dyeing ${his} tail:`, "note");
		}
		frag.append(div);

		div = document.createElement("div");
		div.classList.add("choices");
		div.append(`Colors:`);
		div.append(createList(App.Medicine.Modification.Color.Primary, updatePrimary));
		frag.append(div);

		div = document.createElement("div");
		div.classList.add("choices");
		div.append(`Highlights:`);
		div.append(createList(App.Medicine.Modification.Color.Secondary, updateSecondary));
		frag.append(div);

		if (primaryTailColor !== 0) {
			p = document.createElement("p");
			p.classList.add("choices");
			p.append(
				App.UI.DOM.link(
					`Color ${his} tail`,
					() => {
						slave.tailColor = (primaryTailColor + secondaryTailColor);
						App.Art.refreshSlaveArt(slave, 3, "art-frame");
						if (!cheat) {
							cashX(forceNeg(V.modCost), "slaveMod", slave);
						}
						App.Medicine.Salon.tail(slave); // discard selections after locking them in.
					}
				)
			);
			p.append(` ${primaryTailColor}${secondaryTailColor} now?`);
			frag.append(p);
		}
		return frag;
	}

	function createList(array, method) {
		const links = [];
		for (const item of array) {
			if (item.hasOwnProperty("requirements")) {
				if (item.requirements(slave) === false) {
					continue;
				}
			}
			const title = item.title || capFirstChar(item.value);
			links.push(
				App.UI.DOM.link(
					title,
					() => method(item)
				)
			);
		}
		return App.UI.DOM.generateLinksStrip(links);
	}

	function apply() {
		App.Art.refreshSlaveArt(slave, 3, "art-frame");
		App.Medicine.Salon.tail(
			slave,
			{
				primaryTailColor: primaryTailColor,
				secondaryTailColor: secondaryTailColor,
			}
		);
	}
};
